/* gfClient - A client for the genomic finding program that produces a .psl file. */
#include "common.h"
#include "linefile.h"
#include "fa.h"
#include "genoFind.h"

void gfClient(char *hostName, char *portName, char *nibDir, char *inName, char *outName)
/* gfClient - A client for the genomic finding program that produces a .psl file. */
{
struct lineFile *lf = lineFileOpen(inName, TRUE);
static struct dnaSeq seq;
struct ssBundle *bundleList;
enum ffStringency stringency = ffCdna;
FILE *out = mustOpen(outName, "w");

while (faSpeedReadNext(lf, &seq.dna, &seq.size, &seq.name))
    {
    printf("Processing %s\n", seq.name);
    gfAlignStrand(hostName, portName, nibDir, &seq, FALSE, stringency, 40, gfSavePsl, out);
    reverseComplement(seq.dna, seq.size);
    gfAlignStrand(hostName, portName, nibDir, &seq, TRUE,  stringency, 40, gfSavePsl, out);
    }
printf("Output is in %s\n", outName);
}

void usage()
/* Explain usage and exit. */
{
errAbort(
  "gfClient - A client for the genomic finding program that produces a .psl file\n"
  "usage:\n"
  "   gfClient host port nibDir in.fa out.psl\n");
}

#define TEST_123

#ifdef TEST_123
#include <signal.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include "errabort.h"

static struct sockaddr_in sai;		/* Some system socket info. */

static int setupSocket(char *hostName, char *portName)
/* Set up our socket. */
{
int port;
int sd;
struct hostent *hostent;

if (!isdigit(portName[0]))
    errAbort("Expecting a port number got %s", portName);
port = atoi(portName);
hostent = gethostbyname(hostName);
if (hostent == NULL)
    {
    herror("");
    errAbort("Couldn't find host %s. h_errno %d", hostName, h_errno);
    }
sai.sin_family = AF_INET;
sai.sin_port = htons(port);
memcpy(&sai.sin_addr.s_addr, hostent->h_addr_list[0], sizeof(sai.sin_addr.s_addr));
sd = socket(AF_INET, SOCK_STREAM, 0);
if (sd < 0)
    errnoAbort("Couldn't setup socket %s %s", hostName, portName);
return sd;
}

struct gfRange
/* A range of bases found by genoFind.  Recursive
 * data structure.  Lowest level roughly corresponds
 * to an exon. */
    {
    struct gfRange *next;  /* Next in singly linked list. */
    int qStart;	/* Start in query */
    int qEnd;	/* End in query */
    char *tName;	/* Target name */
    int tStart;	/* Start in target */
    int tEnd;	/* End in target */
    int hitCount;	/* Number of hits */
    struct gfRange *components;	/* Components of range. */
    };

static struct gfRange *rgfRangeLoad(char **row)
/* Load a gfRange from row fetched with select * from gfRange
 * from database.  Dispose of this with gfRangeFree(). */
{
struct gfRange *ret;
int sizeOne,i;
char *s;

AllocVar(ret);
ret->qStart = atoi(row[0]);
ret->qEnd = atoi(row[1]);
ret->tName = cloneString(row[2]);
ret->tStart = atoi(row[3]);
ret->tEnd = atoi(row[4]);
ret->hitCount = atoi(row[5]);
return ret;
}


static void rgfQuerySeq(char *hostName, char *portName, struct dnaSeq *seq)
/* Ask server for places sequence hits. */
{
char buf[256], *row[6];
int rowSize;
struct gfRange *rangeList, *range;

/* Connect to server. */
int sd = setupSocket(hostName, portName);
if (connect(sd, &sai, sizeof(sai)) == -1)
    errnoAbort("Couldn't connect to socket in oneStrand");

/* Do start of query. */
sprintf(buf, "%squery %d", gfSignature(), seq->size);
write(sd, buf, strlen(buf));
read(sd, buf, 1);
if (buf[0] != 'Y')
    errAbort("Expecting 'Y' from server, got %c", buf[0]);
write(sd, seq->dna, seq->size);

/* Read results line by line and save in list, and return. */
for (;;)
    {
    gfRecieveString(sd, buf);
    if (sameString(buf, "end"))
	break;
    else
	{
	uglyf("%s\n", buf);
	rowSize = chopLine(buf, row);
	range = rgfRangeLoad(row);
	slAddHead(&rangeList, range);
	}
    }
close(sd);
}


void gfSimpleClient(char *hostName, char *portName, char *nibDir, char *inName, char *outName)
/* gfClient - A client for the genomic finding program that produces a .psl file. */
{
struct lineFile *lf = lineFileOpen(inName, TRUE);
static struct dnaSeq seq;
struct ssBundle *bundleList;
enum ffStringency stringency = ffCdna;
FILE *out = mustOpen(outName, "w");

while (faSpeedReadNext(lf, &seq.dna, &seq.size, &seq.name))
    {
    printf("Processing %s\n", seq.name);
    rgfQuerySeq(hostName, portName, &seq);
    reverseComplement(seq.dna, seq.size);
    rgfQuerySeq(hostName, portName, &seq);
    }
printf("Output is in %s\n", outName);
}
#endif /* TEST_123 */


int main(int argc, char *argv[])
/* Process command line. */
{
if (argc != 6)
    usage();
gfClient(argv[1], argv[2], argv[3], argv[4], argv[5]);
// gfSimpleClient(argv[1], argv[2], argv[3], argv[4], argv[5]);
return 0;
}
